//
//  RootViewController.m
//  FTS
//
//  Created by Michael Heyeck on 2/4/10.
//  Copyright Fair Oaks Labs, Inc. 2010. All rights reserved.
//

#import "Text.h"
#import "Keyword.h"
#import "Tokenizer.h"
#import "TextEdit.h"
#import "RootViewController.h"


@interface RootViewController ()

@property (nonatomic, retain) LoadController* loadController;
@property (nonatomic, retain, readonly) NSFetchedResultsController* textResultsController;
@property (nonatomic, retain, readonly) NSFetchedResultsController* keywordResultsController;

- (void)flushMOC;

- (UITableViewCell*)tableView:(UITableView*)tableView textCellForRowAtIndexPath:(NSIndexPath*)indexPath;
- (void)tableView:(UITableView*)tableView didSelectTextRowAtIndexPath:(NSIndexPath*)indexPath;

- (UITableViewCell*)tableView:(UITableView*)tableView keywordCellForRowAtIndexPath:(NSIndexPath*)indexPath;
- (void)tableView:(UITableView*)tableView didSelectKeywordRowAtIndexPath:(NSIndexPath*)indexPath;

@end


@implementation RootViewController

@synthesize useAll;
@synthesize keywords;
@synthesize managedObjectContext;

- (LoadController*)loadController
{
	return [[loadController retain] autorelease];
}

- (void)setLoadController:(LoadController*)newLoadController
{
	if (loadController != newLoadController)
	{
		loadController.target = nil;
		[loadController release];
		
		loadController = [newLoadController retain];
	}
}

- (NSFetchedResultsController*)textResultsController
{
	if (!textResultsController)
	{
		NSFetchRequest* fetchRequest = [[[NSFetchRequest alloc] init] autorelease];
		
		// Entities to retrieve
		fetchRequest.entity = [NSEntityDescription entityForName:@"Text" inManagedObjectContext:self.managedObjectContext];
		
		// Batch them
		fetchRequest.fetchBatchSize = 20;
		
		// Configure a predicate, if appropriate
		if ([self.keywords count])
		{
			if (useAll)
			{
				NSMutableString* s = [NSMutableString stringWithString:@"(ANY words = %@)"];
				for (NSInteger i = [self.keywords count]-2; i >= 0; i--)
				{
					[s appendString:@" AND (ANY words = %@)"];
				}
				fetchRequest.predicate = [NSPredicate predicateWithFormat:s argumentArray:[self.keywords allObjects]];
			}
			else
			{
				fetchRequest.predicate = [NSPredicate predicateWithFormat:@"ANY words IN %@",self.keywords];
			}
		}
		
		// Sort them
		NSSortDescriptor* sortDescriptor = [[NSSortDescriptor alloc] initWithKey:@"sortPrefix" ascending:YES selector:@selector(localizedCompare:)];
		fetchRequest.sortDescriptors = [NSArray arrayWithObject:sortDescriptor];
		[sortDescriptor release];
		
		// Create and configure NSFRC
		textResultsController = [[NSFetchedResultsController alloc] initWithFetchRequest:fetchRequest
																	managedObjectContext:self.managedObjectContext
																	  sectionNameKeyPath:nil
																			   cacheName:nil];
		textResultsController.delegate = self;
	}
	return textResultsController;
}

- (NSFetchedResultsController*)keywordResultsController
{
	if (!keywordResultsController)
	{
		NSFetchRequest* fetchRequest = [[[NSFetchRequest alloc] init] autorelease];
		
		// Entities to retrieve
		fetchRequest.entity = [NSEntityDescription entityForName:@"Keyword" inManagedObjectContext:self.managedObjectContext];
		
		// Batch them
		fetchRequest.fetchBatchSize = 20;
		
		// Sort them
		NSSortDescriptor* sortDescriptor = [[NSSortDescriptor alloc] initWithKey:@"keyword" ascending:YES selector:@selector(localizedCompare:)];
		fetchRequest.sortDescriptors = [NSArray arrayWithObject:sortDescriptor];
		[sortDescriptor release];
		
		// Create and configure NSFRC
		keywordResultsController = [[NSFetchedResultsController alloc] initWithFetchRequest:fetchRequest
																	   managedObjectContext:self.managedObjectContext
																		 sectionNameKeyPath:nil
																				  cacheName:nil];
	}
	return keywordResultsController;
}


- (void)viewDidLoad
{
    [super viewDidLoad];

	if (self.keywords == nil)
	{
		// TL controller
		self.title = @"FTS Demo";
		self.navigationItem.backBarButtonItem = [[[UIBarButtonItem alloc] initWithTitle:@"Home" style:UIBarButtonItemStyleBordered target:nil action:nil] autorelease];

		self.navigationItem.rightBarButtonItem = [[[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemAdd target:self action:@selector(insertNewObject)] autorelease];
		
		[self.searchDisplayController setActive:YES animated:NO];
		[self.searchDisplayController setActive:NO animated:NO];
	}
	else
	{
		// Filtered controller
		self.title = @"Search Results";
	}

	NSError* error = nil;
	if (![self.textResultsController performFetch:&error])
	{
		NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
	}
}


/*
- (void)viewWillAppear:(BOOL)animated {
    [super viewWillAppear:animated];
}
*/
- (void)viewDidAppear:(BOOL)animated
{
    [super viewDidAppear:animated];

	if (self.keywords)
	{
		// Filtered controller
		if (![self tableView:self.tableView numberOfRowsInSection:0])
		{
			[[[[UIAlertView alloc] initWithTitle:nil
										 message:@"No records matched this search"
										delegate:self
							   cancelButtonTitle:@"Ok"
							   otherButtonTitles:nil] autorelease] show];
		}
	}
}
/*
- (void)viewWillDisappear:(BOOL)animated {
	[super viewWillDisappear:animated];
}
*/
/*
- (void)viewDidDisappear:(BOOL)animated {
	[super viewDidDisappear:animated];
}
*/

/*
// Override to allow orientations other than the default portrait orientation.
- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation {
    // Return YES for supported orientations
    return (interfaceOrientation == UIInterfaceOrientationPortrait);
}
*/

- (void)didReceiveMemoryWarning {
    [super didReceiveMemoryWarning]; // Releases the view if it doesn't have a superview
    // Release anything that's not essential, such as cached data
}


- (void)insertNewObjects
{
	self.loadController = [[[LoadController alloc] initWithNibName:@"LoadController" bundle:nil] autorelease];
	[self.loadController startLoadWithTarget:self action:@selector(flushMOC)];
}


- (void)insertNewObject
{
	TextEdit* te = [[[TextEdit alloc] initWithNibName:@"TextEdit" bundle:nil] autorelease];
	
	te.text = nil;
	te.managedObjectContext = self.managedObjectContext;
	
	[self.navigationController pushViewController:te animated:YES];
}


- (void)dealloc
{
	self.loadController = nil;
	[textResultsController release];
	[keywordResultsController release];
	[keywords release];
	[managedObjectContext release];
    [super dealloc];
}

#pragma mark Table view data source methods

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView {
    return 1;
}


// Customize the number of rows in the table view.
- (NSInteger)tableView:(UITableView*)tableView numberOfRowsInSection:(NSInteger)section
{
	NSArray* sections;
	if (tableView == self.tableView)
	{
		// Primary table view
		sections = [self.textResultsController sections];
	}
	else
	{
		// Search results table view
		sections = [self.keywordResultsController sections];
	}

	if ([sections count])
	{
		return [[sections objectAtIndex:section] numberOfObjects];
	}
	else
	{
		return 0;
	}
}


// Customize the appearance of table view (text block) cells.
- (UITableViewCell*)tableView:(UITableView*)tableView textCellForRowAtIndexPath:(NSIndexPath*)indexPath
{
    static NSString* CellIdentifier = @"Text";
    
    UITableViewCell* cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
    if (cell == nil)
	{
        cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:CellIdentifier] autorelease];
		cell.accessoryType = UITableViewCellAccessoryNone;
    }
    
    cell.textLabel.text = [[self.textResultsController objectAtIndexPath:indexPath] text];
    return cell;
}


// Customize the appearance of table view (keyword) cells.
- (UITableViewCell*)tableView:(UITableView*)tableView keywordCellForRowAtIndexPath:(NSIndexPath*)indexPath
{
    static NSString* CellIdentifier = @"Keyword";
    
    UITableViewCell* cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
    if (cell == nil)
	{
        cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:CellIdentifier] autorelease];
		cell.accessoryType = UITableViewCellAccessoryNone;
    }
	
    cell.textLabel.text = [[self.keywordResultsController objectAtIndexPath:indexPath] keyword];
    return cell;
}


// Customize the appearance of table view cells.
- (UITableViewCell*)tableView:(UITableView*)tableView cellForRowAtIndexPath:(NSIndexPath*)indexPath
{
	if (tableView == self.tableView)
	{
		return [self tableView:tableView textCellForRowAtIndexPath:indexPath];
	}
	else
	{
		return [self tableView:tableView keywordCellForRowAtIndexPath:indexPath];
	}
}


// Override to support conditional editing of the table view.
- (BOOL)tableView:(UITableView*)tableView canEditRowAtIndexPath:(NSIndexPath*)indexPath
{
	// Only the primary table view is editable
	return (tableView == self.tableView);
}


// Override to support editing the table view.
- (void)tableView:(UITableView*)tableView commitEditingStyle:(UITableViewCellEditingStyle)editingStyle forRowAtIndexPath:(NSIndexPath*)indexPath
{
	// Only the primary table view is editable
	if (tableView != self.tableView) return;
	
    if (editingStyle == UITableViewCellEditingStyleDelete)
	{
        // Delete the managed object for the given index path
		[self.managedObjectContext deleteObject:[self.textResultsController objectAtIndexPath:indexPath]];
		
		// Save the context.
		NSError* error = nil;
		if (![self.managedObjectContext save:&error])
		{
			NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
		}
	}   
}


/*
// Override to support rearranging the table view.
- (void)tableView:(UITableView *)tableView moveRowAtIndexPath:(NSIndexPath *)fromIndexPath toIndexPath:(NSIndexPath *)toIndexPath {
}
*/


/*
// Override to support conditional rearranging of the table view.
- (BOOL)tableView:(UITableView *)tableView canMoveRowAtIndexPath:(NSIndexPath *)indexPath {
    // Return NO if you do not want the item to be re-orderable.
    return YES;
}
*/

#pragma mark Table view delegate methods

- (void)tableView:(UITableView*)tableView didSelectTextRowAtIndexPath:(NSIndexPath*)indexPath
{
	TextEdit* te = [[[TextEdit alloc] initWithNibName:@"TextEdit" bundle:nil] autorelease];
	
	te.text = [self.textResultsController objectAtIndexPath:indexPath];
	te.managedObjectContext = self.managedObjectContext;
	
	[self.navigationController pushViewController:te animated:YES];
}


- (void)tableView:(UITableView*)tableView didSelectKeywordRowAtIndexPath:(NSIndexPath*)indexPath
{
	NSString* searchString = self.searchDisplayController.searchBar.text;
	
	NSInteger position;
	[[Tokenizer sharedTokenizer] findLastToken:searchString position:&position];
	
	if (position != NSNotFound)
	{
		NSString* keyword = [[self.keywordResultsController objectAtIndexPath:indexPath] keyword];
		NSString* newString = [[searchString substringToIndex:position] stringByAppendingFormat:@"%@ ",keyword];
		self.searchDisplayController.searchBar.text = newString;
	}
	else
	{
		[tableView deselectRowAtIndexPath:indexPath animated:YES];
	}
	
	[self.searchDisplayController.searchBar becomeFirstResponder];
}


- (void)tableView:(UITableView*)tableView didSelectRowAtIndexPath:(NSIndexPath*)indexPath
{
	if (tableView == self.tableView)
	{
		return [self tableView:tableView didSelectTextRowAtIndexPath:indexPath];
	}
	else
	{
		return [self tableView:tableView didSelectKeywordRowAtIndexPath:indexPath];
	}
}


- (NSIndexPath*)tableView:(UITableView*)tableView willSelectRowAtIndexPath:(NSIndexPath*)indexPath
{
	return indexPath;
}

#pragma mark Fetched results controller delegate methods

- (void)controllerDidChangeContent:(NSFetchedResultsController*)controller
{
	// Simple (if slightly inelegant) handler: just reload the table view
	[self.tableView reloadData];
}

#pragma mark Alert view delegate methods

- (void)alertView:(UIAlertView*)alertView clickedButtonAtIndex:(NSInteger)buttonIndex
{
	if (self.keywords == nil)
	{
		// TL controller
		[self.searchDisplayController.searchBar becomeFirstResponder];
	}
	else
	{
		// Filtered controller
		[self.navigationController popViewControllerAnimated:YES];
	}
}

#pragma mark Search bar delegate methods

- (void)searchBarSearchButtonClicked:(UISearchBar*)searchBar
{
	// Process the search string for keywords
	NSMutableSet* tokens = [[Tokenizer sharedTokenizer] tokenize:searchBar.text];
	
	// Create and perform a request for existing keyword records for these tokens
	NSError* err;
	NSFetchRequest* request = [[NSFetchRequest alloc] init];
	request.entity = [NSEntityDescription entityForName:@"Keyword" inManagedObjectContext:self.managedObjectContext];
	request.predicate = [NSPredicate predicateWithFormat:@"keyword IN %@",tokens];
	NSArray* aKeywords = [self.managedObjectContext executeFetchRequest:request error:&err];
	[request release];

	// Check for trivially empty sets
	if (![aKeywords count])
	{
		[[[[UIAlertView alloc] initWithTitle:nil
									 message:@"No records match this search; it contains no indexed keywords"
									delegate:self
						   cancelButtonTitle:@"Ok"
						   otherButtonTitles:nil] autorelease] show];
		return;
	}
	if (searchBar.selectedScopeButtonIndex && ([tokens count] != [aKeywords count]))
	{
		[[[[UIAlertView alloc] initWithTitle:nil
									 message:@"No records match this search; it contains one or more keywords with no matches"
									delegate:self
						   cancelButtonTitle:@"Ok"
						   otherButtonTitles:nil] autorelease] show];
		return;
	}
	
	// Create a filtered result set
	RootViewController* vc = [[[RootViewController alloc] initWithNibName:@"ResultViewController" bundle:nil] autorelease];
	vc.managedObjectContext = self.managedObjectContext;
	vc.keywords = [NSSet setWithArray:aKeywords];
	vc.useAll = searchBar.selectedScopeButtonIndex?YES:NO;
	
	// Exit search mode
	[self.searchDisplayController setActive:NO animated:YES];
	
	// Present the result set
	[self.navigationController pushViewController:vc animated:YES];
}

#pragma mark Search display delegate methods

- (BOOL)searchDisplayController:(UISearchDisplayController*)controller shouldReloadTableForSearchScope:(NSInteger)searchOption
{
	return NO;
}


- (BOOL)searchDisplayController:(UISearchDisplayController*)controller shouldReloadTableForSearchString:(NSString*)searchString
{
	NSString* start = [[Tokenizer sharedTokenizer] findLastToken:searchString position:nil];
	if (start)
	{
		NSString* stop = [start stringByAppendingString:@"zzz"];
		NSArray* range = [NSArray arrayWithObjects:[NSExpression expressionForConstantValue:start],[NSExpression expressionForConstantValue:stop],nil];
		self.keywordResultsController.fetchRequest.predicate = [NSPredicate predicateWithFormat:@"keyword BETWEEN %@",range];
	}
	else
	{
		self.keywordResultsController.fetchRequest.predicate = [NSPredicate predicateWithValue:NO];
	}
	
	NSError* err;
	[self.keywordResultsController performFetch:&err];
	
	return YES;
}

#pragma mark Extension methods

- (void)flushMOC
{
	// A pretty blatant hack (to avoid wasting memory)
	self.loadController = nil;

	// This is very dangerous; no objects from this MOC may be retained anywhere else in the code
	[self.managedObjectContext reset];

	NSError* err;
	[self.textResultsController performFetch:&err];
	
	[self.tableView reloadData];
}

@end

